from dune.perftool.generation import (cached,
                                      get_counted_variable,
                                      global_context,
                                      include_file,
                                      preamble,
                                      )
from dune.perftool.pdelab.driver import (FEM_name_mangling,
                                         get_formdata,
                                         get_trial_element,
                                         is_stationary,
                                         preprocess_leaf_data,
                                         )
from dune.perftool.pdelab.driver.gridfunctionspace import (name_trial_gfs,
                                                           name_leafview,
                                                           )
from dune.perftool.pdelab.driver.gridoperator import (name_parameters,)

from ufl import FiniteElement, MixedElement, TensorElement, VectorElement, TensorProductElement


def _do_interpolate(dirichlet):
    if isinstance(dirichlet, (list, tuple)):
        return any(bool(d) for d in dirichlet)
    else:
        return bool(dirichlet)


def interpolate_dirichlet_data(name):
    element = get_trial_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    if _do_interpolate(is_dirichlet) or not is_stationary():
        gfs = name_trial_gfs()
        dirichlet = preprocess_leaf_data(element, "dirichlet_expression")
        bf = name_boundary_function(element, dirichlet)
        interpolate_vector(bf, gfs, name)


@preamble
def interpolate_vector(func, gfs, name):
    return "Dune::PDELab::interpolate({}, {}, {});".format(func,
                                                           gfs,
                                                           name,
                                                           )


@cached
def name_boundary_function(element, func):
    if isinstance(element, MixedElement):
        k = 0
        childs = []
        for subel in element.sub_elements():
            childs.append(name_boundary_function(subel, func[k:k + subel.value_size()]))
            k = k + subel.value_size()
        name = "_".join(childs)
        if len(childs) == 1:
            name = "{}_dummy".format(name)
        define_compositegfs_parameterfunction(name, tuple(childs))
        return name
    else:
        assert isinstance(element, (FiniteElement, TensorProductElement))
        name = get_counted_variable("func")
        define_boundary_function(name, func[0])
        return name


@preamble
def define_compositegfs_parameterfunction(name, children):
    return "Dune::PDELab::CompositeGridFunction<{}> {}({});".format(', '.join('decltype({})'.format(c) for c in children),
                                                                    name,
                                                                    ', '.join(children))


@preamble
def define_boundary_function(name, dirichlet):
    gv = name_leafview()
    lambdaname = name_boundary_lambda(dirichlet)
    include_file('dune/pdelab/function/callableadapter.hh', filetag='driver')
    if is_stationary():
        return "auto {} = Dune::PDELab::makeGridFunctionFromCallable({}, {});".format(name,
                                                                                      gv,
                                                                                      lambdaname,
                                                                                      )
    else:
        params = name_parameters(get_formdata())
        return "auto {} = Dune::PDELab::makeInstationaryGridFunctionFromCallable({}, {}, {});".format(name,
                                                                                                      gv,
                                                                                                      lambdaname,
                                                                                                      params)


@cached
def name_boundary_lambda(boundary):
    name = get_counted_variable("lambda")
    define_boundary_lambda(name, boundary)
    return name


@preamble
def define_boundary_lambda(name, boundary):
    if boundary is None:
        boundary = 0.0
    from ufl.classes import Expr
    if isinstance(boundary, (int, float)):
        return "auto {} = [&](const auto& x){{ return {}; }};".format(name, float(boundary))
    elif isinstance(boundary, Expr):
        from dune.perftool.loopy.target import type_floatingpoint
        from dune.perftool.pdelab.driver.visitor import ufl_to_code
        return "auto {} = [&](const auto& x){{ return ({}){}; }};".format(name,
                                                                          type_floatingpoint(),
                                                                          ufl_to_code(boundary))
    else:
        raise NotImplementedError("What is this?")
